.TL
Using the Bulk Loading Tools to Support a Directory Service
.AU
John Colville
.AI
School of Computing Sciences
University of Technology, Sydney
.FS
.nh
Author's e-mail address: colville@socs.uts.edu.au
.br
This work discussed in this paper was conducted while the author was visiting
Computing and Media Services,
Brunel University, United Kingdom
.FE
.sp 1
.NH 1
Introduction
.nh
.PP
The Bulk Loading Tools (BLT), \fBbulkload\fR and \fBnewdmdiff\fR, [1] are
designed
to help with loading and managing an X.500 Directory Service [2]. Both
tools are designed to automate the manipulation of large quantities of data.
\fB\fBBulkload\fR\fR up-loads data from multiple sources into the directory.
It has a
subfunction, bulk management, which checks consistency between the directory
and data from a number of data sources.
\fBNewdmdiff\fR generates a list of the
differences between two sets of data. This is designed to make it easier to
perform incremental updates on the directory.
.PP
This paper provides some suggestions and methology for using these tools
effectively. A more complete description of BLT is provided in [1].
The major part of this paper describes how an administrator can use BLT to:
.PP
.IP \(bu
Set up a directory service
.IP \(bu
Support continuing updates to the directory
.IP \(bu
Ease the conversion from
a currently operating directory to one which takes advantage of BLT.
.PP
Decisions as to the structure of The Directory and the names to use within its
structure, are beyond the scope of this paper. There are a number of papers
which discuss these topics [3 - 5]. This paper also assumes that you
already know how to create a minimal DSA with \fIQUIPU\fR [6].
.PP
The next section
provides a brief overview of the stages of creating and
maintaining a directory using BLT.
After a discussion of the files used by \fBbulkload\fR, and their formats,
this paper describes, in greater detail,
the steps in setting up and loading a new directory. It
describes the algorithms which allow \fBbulkload\fR to match information from
different sources. A number of \fBperl\fR scripts to help with this process, are
provided as Appendixes.
.NH 1
Creating and Maintaining a Directory with BLT
.PP
.nh
To set up a directory service you need to go through the following steps.
Many of them are elaborated as later sections of this paper as indicated by the
Section numbers in square brackets at the end of the step.
.IP 1.
Obtain the data from the various sources around the organisation e.g.
staff names and affiliations from personnel records, telephone lists,
e-mail addresses, etc. 
It is preferable if this information can be delivered to you in machine
readable form: even better if it can be obtained automatically.
It is better still if you can arrange to receive regular updates.
.IP 2.
Design your directory structure and decide which data sources are to `master'
each attribute of each entry.
.IP 3.
Write conversion tools to get the data into a format suitable for entry into
\fBbulkload\fR
[Sect. 8 and Appendixes]
.IP 4.
Create the structure of your directory [Sects. 6, 7].
.IP 5.
Load each data source in turn, starting with the most authoritative ones
[Sect. 11].
.IP 6.
On a regular basis, get updated data. Compare it with previous data using
\fBnewdmdiff\fR. Then apply updates using \fBbulkload\fR [Sect. 12].
.NH 1
Files Associated with Bulkload
.PP
.nh
\fBBulkload\fR uses two input files and produces a number of output files.
It starts
by reading a file, \fBbulkloadtailor\fR,
which defines the configuration for this
run.  By default is uses the file
.IP
$(ETCDIR)bulk/bulkloadtailor
.LP
where $ETCDIR is the \fB/etc\fR directory which has been set up when compiling
\fIQUIPU\fR.
An alternative configuration file can be used by putting
.IP
-t \fIalternativetailorfile\fR
.LP
in the command line. This may be useful when you have a number of different
tailor files for different passes through the data.
.PP
\fBBulkload\fR then invites you to operate in non-interactive mode or
interactively. (Most of the time, we would prefer to operate in non-interactive
mode, but inevitably a fraction must be dealt with interactively.) After that,
\fBbulkload\fR reads its data from the file you specify in the command line.
This data file shoudl be in \fIdmformat\fR when the directory is being loaded
with new data, or in \fInewdmformat\fR during incremental updates.
.PP
\fBBulkload\fR reports to stdout the RDN of each entry that it is updating.
It also generates three journal files. Settings within the bulkloadtailor file
decide which output goes to each of these files.
The three journal files are designed to hold lists of
changes which have been made, ignored as unnecessary
(e.g. the addition of an attribute value that already exists),
and ignored as `too hard'
(e.g. updates that can not be resolved automatically).
These files are given the name of the input file suffixed by .done, .noAction
and .PROBLEM respectively. So, for example, if the input came from file dmdata,
the journal files would be dmdata.done, dmdata.noAction and dmdata.PROBLEM.
.PP
The journal files are produced in formats which are valid as input to
\fBbulkload\fR. This can be useful where, for example, one run is used to add new
entries and to update unambiguously identified entries. The .PROBLEM file is
then input to another run, with different tailoring, to update the other
entries.
.NH 1
The Bulkloadtailor File
.PP
.nh
The bulkloadtailor file defines the options that are to be used by a
particular run of \fBbulkload\fR.
Most of the parameters are self-explanatory, with
careful reading. One group specifies which sources of data are to be regarded as
authoritative for the various attributes.
This is discussed in more detail in Section 7.
.PP
Another group determines which of the
different journal files receive the reports on each action. If you expect
that different data sources will provide the same attribute but different
levels of reliablility, it is worth setting up different bulkloadtailor
files to reflect these differences. Then the relevant tailoring file can be used
to control the distribution of journal messages for each  input file.
.NH 1
Input Formats for Bulkload
.PP
.nh
\fBBulkload\fR accepts data in two formats: \fIdmformat\fR and
\fInewdmformat\fR.
A file in \fIdmformat\fR looks like \fIQUIPU's Entry Data Block\fR file [6],
except that it
includes \fBsource\fR and \fBrootedAt\fR statements.
The \fBsource\fR line names the data source.
The \fBrootedAt\fR statement defines the parent directory for the
entries which follow it.
All the data supplied by a single
source can be included in a single file,
with sets of data for different DIT
subtrees each being preceded by a \fBrootedAt\fR statement.
Each \fBsource\fR and \fBrootedAt\fR statement and
the entries are separated by one or more blank lines. As with EDB files,
lines starting with # are treated as comment lines.
.KS
.SH
Example of a dmformat file
.ft C
.B
.TS
box;
l.
source= adminStaff

rootedAt=c=GB@o=Brunel University@ou=Computing and Media Services

cn= A J Findlay
cn= Andrew Findlay
personalTitle= Dr
surname= Findlay
organizationalStatus= Staff
objectClass= top & person & organizationalPerson & pilotObject & newPilotPerson & quipuObject

cn= D Mahl
cn= Damanjit Mahl
personalTitle= Mr
surname= Mahl
organizationalStatus= Staff
objectClass= top & person & organizationalPerson & pilotObject & newPilotPerson & quipuObject

rootedAt=c=GB@o=Brunel University@ou=Computer Science

cn= D Martland
cn= David Martland
personalTitle= Dr
surname= Martland
organizationalStatus= Staff
objectClass= top & person & organizationalPerson & pilotObject & newPilotPerson & quipuObject

# etc. ....
.TE
.ft
.KE
.R
.PP
\fINewdmformat\fR provides a list of changes to be made during an incremental
update. It is most easily generated by using \fBnewdmdiff\fR.
A simple \fInewdmformat\fR
file to change A J Findlay's organizationalStatus from Staff to Student, is
shown below:
.ft C
.B
.TS
box;
l.
newdmformat

source= adminStaff

rootedAt=c=GB@o=Brunel University@ou=Computing and Media Services

cn= A J Findlay
<organizationalStatus= Staff
>organizationalStatus= Student
.TE
.ft
.R
.NH 1
Setting Up a New Directory with BLT
.PP
.nh
The directory is organised as a tree of entries with the national and
supra-national entities at the top. Most organisations will fit directly below
the entry for their country.  To start with, you must decide what is the
relevant structure to reflect your organisation to the outside world. You must
also decide on the scheme to be used for names within your organisation [3].
.PP
When designing your organisational structure, do not be afraid to experiment
with alternative structures until you find a satisfactory one.  It is useful to
make a series of \fItar\fR files of the contents of the directory,
as you add more
data. Then, if you decide to change the structure,  it is easy to roll the
directory back to an earlier, intermediate version.
.PP
Before you start to use the BLT, you must create a minimal directory 
containing the entries down the tree to, and including, your organisation
entry.  It must also contain an entry for your Directory Manager.
.PP
You will probably find that the data comes from a number of different sources.
It is extremely unlikely that the data formats will be compatible between these
sources (or even within a single source). These different formats can be
massaged into more compatible styles by use of \fBawk\fR or \fBperl\fR.
Some \fBperl\fR scripts are provided in the Appendixes to this document.
.PP
These \fBperl\fR scripts include \fBmake-ou-load-file\fR [Appendix 6]
which generates a set of OU entries
from a set of files containing leaf node entries. So you can produce your leaf
node \fIdmformat\fR files and generate OUs consistent with them.
.PP
Often there are some attributes which are the same for all entries e.g.
objectClass attributes.
Rather than embedding long output lines in each script,
these fixed attributes are probably best prepared as
a separate file which is given as data to the script. 
For example see the `boilerplate' files in Appendix 7.
.PP
Unless great care is taken in processing the data from the various sources, the
directory will readily acquire inconsistent data. Here, too, the use of tar
files makes it easy to recover from updating disasters.
.NH 1
Loading the BLT Data Mnagement Infrastructure
.PP
.nh
When you have identified the data sources that are providing your data, you must
tell the directory
which attributes each source can provide,
the `bulk data management infrastructure'. This
is most easily accomplished by giving \fBbulkload\fR a file that defines this
information. Each source becomes an entry of the special type dataSource. It is
useful to create an extra dataSource `directoryStructure' which controls the
loading of the organisational structure. This allows you to set up the
directory tree using \fBbulkload\fR.
.PP
The entry for each dataSource specifies the attributes
for which it is a `master source'
i.e. the attributes for which it is regarded as authoritative.
The entry also defines whether it can `create'
extra entries or `delete' entries. The following options from the
bulkloadtailor file determine how authoritatively to regard each attribute
supplied by each dataSource:
.IP
.nf
When no master specified for a type all sources are masters
All sources are masters for
.fi
.LP
For example, using:
.IP
.nf
When no master specified for a type all sources are masters : YES
.fi
.LP
This means that \fBbulkload\fR regards every attribute type which is not
specifically mentioned as having a particular mastering source,
as being authoritative from every source. Conversely, you
can produce a list of exactly those attributes for which all sources can be
treated as authoritative by using, for example:
.ft C
.B
.TS
box;
l.
When no master specified for a type all sources are masters : NO
All sources are masters for : objectClass
All sources are masters for : acl
All sources are masters for : favouriteDrink
.TE
.ft
.KE
.R
.bp
An infrastructure file is:
.ft C
.B
.TS
box;
l.
dataSource

source=directoryStructure

rootedAt=c=GB@o=Brunel University

dataSource=directoryStructure
masters=treeStructure
masters=masterDSA
deleteSource=TRUE
createSource=TRUE

dataSource=adminStaff
masters=organizationalStatus
masters=personalTitle
masters=surname
deleteSource=TRUE
createSource=TRUE

dataSource=adminStudents
masters=organizationalStatus
masters=personalTitle
masters=surname
deleteSource=TRUE
createSource=TRUE

dataSource=emailData
masters=rfc822Mailbox
masters=uid
deleteSource=TRUE
createSource=TRUE

dataSource=telephoneData
masters=telephoneNumber
masters=roomNumber
deleteSource=TRUE
createSource=TRUE
.TE
.ft
.KE
.R
.PP
This example assumes that each of the dataSources is going to provide some
extra attributes and may provide some new entries. For example the dataSources
adminStaff and adminStudents both provide information about people from the
organisation's administrative database. Although these inputs may come from the same place, it is worth keeping these sources
separate because they may be delivered in different formats
and require different preprocessing. The dataSource emailData provides only the
userid and the mailbox address of people whose entries have already been loaded.
However, it can also be used to add extra entries for people
who are in neither
the staff nor the student database.
These entries can include people like
honorary staff who do not appear in the data derived from the administrative
payroll-based system.
A more extensive example of an infrastructure file is provided in [1].
.PP
Having loaded the infrastructure information you can then populate the directory
at that level. It is sensible to perform one run in which you add only the
OUs at this level.
This can be accomplished by providing \fBbulkload\fR with a file
which contains the OU entries.
.bp
.ft C
.B
.TS
box;
l.
source= directoryStructure
 
rootedAt=c=GB@o=Brunel University
 
ou=Administration
masterDSA=c=GB@cn=basilisk
treeStructure= alias & organizationalUnit & organizationalPerson & organizationalRole & applicationProcess
treeStructure= pilotObject & newPilotPerson & room & quipuObject &quipuNonLeafObject
treeStructure= bulkDataSourceEntry & bulkDataSourceDetails & bulkDataEntryDetails & mhsOrganizationalUser
objectClass= top & organizationalUnit & pilotObject & quipuObject & quipuNonLeafObject
 
ou=Computing and Media Services
masterDSA=c=GB@cn=basilisk
treeStructure= alias & organizationalUnit & organizationalPerson & organizationalRole & applicationProcess
treeStructure= pilotObject & newPilotPerson & room & quipuObject &quipuNonLeafObject
treeStructure= bulkDataSourceEntry & bulkDataSourceDetails & bulkDataEntryDetails & mhsOrganizationalUser
objectClass= top & organizationalUnit & pilotObject & quipuObject & quipuNonLeafObject

# etc ...
.TE
.ft
.KE
.R
.PP
This example will load the entries for ou=Administration and
ou=Computing and Media Services.
.PP
If you have more than one level of OU within your organisation, it is important
to realise that you can only add `infrastructural' entries one level at a time.
In general sub-OUs need to be set up in the following stages:
.IP
.nf
Load dataSource information for one level
Create OUs at that level
Load dataSource information for lower level
Create OUs at lower level
	etc.
.fi
.LP
For example, after the run which added ou=Computing and Media Services, we could
use the following extra `infrastructure' file to define the sub-OU External
Computer Users:
.ft C
.B
.TS
box;
l.
source=directoryStructure
 
rootedAt=c=GB@o=Brunel University@ou=Computing and Media Services@ou=External Computer Users
 
dataSource=emailData
masters=organizationalStatus
masters=personalTitle
masters=surname
masters=rfc822Mailbox
masters=uid
deleteSource=TRUE
createSource=TRUE
.TE
.ft
.R
.NH 1
Preparing the Leaf Entry Data
.PP
.nh
A major problem is massaging your data into a form acceptable to BLT. Of course,
each organisation and each data source will be different. However, certain
preparation processes can be automated in similar ways, no matter what the site.
.PP
Data from some sources may be provided in different actual organisational
units from those that are wanted in the directory implementation. BLT provides
the mechanism of sub-directory search lists. These allow data which one 
source treats as from a single OU to be split into a number of sub-OUs.
A sub-directory search list specifies a sub-tree parent entry together
with a list of distinguished name extensions.
.PP
.bp
For example,
consider the following Directory structure:-
.IP
.RS
.IP
.so dir.pic
.RE
.LP
.PP
If the Registry telephone data contained information about ou=Accounts,
ou=Exams and ou=Records,
but did not indicate which entries should go within each of these departments,
the required bulkSubDirList \fIentry\fR would be :-
.ft C
.B
.TS
box;
l.
bulkSubDirList= RegistryList
bulkSubTreeDN= c=GB@o=Brunel
bulkSubTreeDNextension= ou=Accounts
bulkSubTreeDNextension= ou=Exams
bulkSubTreeDNextension= ou=Records
.TE
.ft
.R
.PP
which would be placed in the DIT at c=GB@o=Brunel.
The special attribute value \fBbulkSubDirList=All\fR means that updates will be
applied to all sub-OUs within the current OU.
.PP
The reverse problem i.e. merging more than one nominal OU onto a single OU
in the directory, cannot be handled directly in \fBbulkload\fR.
So, \fBbulkload\fR would not be of immediate use,
if we wanted to `flatten' the directory of
the previous example, by merging the three divisions ou=Accounts,
ou=Exams and ou=Records, into a single ou=Registry.

Such a flattening operation would have to be applied to the data before
presenting it to \fBbulkload\fR.
The \fBperl\fR script,
\fBmake-phone-info-from-list\fR [Appendix 2] produces a file of entries,
performing this merging of OU groups as desired. It also outputs the original
ou as an attribute within each cn which is part of a merged OU. This means
that the original OU can still be found by DUAs which have the right
intelligence. The output from \fBmake-phone-info-from-list\fR can be then used
as input to \fBbulkload\fR.
.NH 1
Adding Leaf Entries to the Directory
.PP
.nh
Having loaded the infrastructure, it is then a simple, if tedious, process
to add all of the personal and other entries using files similar to the
\fIdmformat\fR example given earlier. These files could, of course, be prepared
by using \fBperl\fR scripts similar to those in the Appendixes.
.PP
It is sensible to attempt to add all of the data from one data source before
attempting to add data from another source. In that way, any anomalies with
the data from one source, can be fixed without affecting the other data.
Addition of all the data from one source may involve more than one pass over
the data. A suggested method of performing these updates is described
in Section 11.
.PP
As the updates proceed, the contents of the directory can be monitored using
\fBdish\fR or another, more friendly, DUA. For example, within \fBdish\fR,
using the \fB-all\fR
read flag, will show all of the attributes of an entry, including some which are
not displayed by more user-oriented DUAs. The \fB-all\fR flag will,
for example, display the values of the algoName attribute
which is generated by \fBbulkload\fR (AlgoName
is discussed in Section 10).
.PP
It is important to ensure that the DUA is talking directly to the \fIlocal\fR
DSA rather than to a `shadowed' copy. Also you must read
the latest version of an entry
by overriding any cacheing of entries which may have taken place.
If inspecting the DSA using \fBdish\fR
this is guaranteed by using the \fB-dontusecopy\fR service control flag.
.PP
Do not discard the leaf entry dmformat file from each data source after you
have loaded its data. When you have later changes to make to the directory,
this file can be used by \fBnewdmdiff\fR to prepare a file of updates.
Applying the only update file to the directory will speed up these updates.
.NH
Matching Entries
.PP
.nh
Data from different sources will frequently identify the same person by
different purported RDNs e.g. one source may identify a person as Paul Barker
whereas another source calls him P F Barker. \fBBulkload\fR provides two ways of
correlating entries: exact matching and algorithmic matching. Exact matching is
the default mode of operation.
.PP
For each dataSource and RDN, once a match has been made, \fBbulkload\fR puts an
algoName attribute in the entry which lists the correspondence between that RDN
and an entry/record in the dataSource.
.PP
With exact matching, the program will match data from different sources only if
the RDNs are exactly the same in the two dataSources.
.PP
\fBBulkload\fR uses two variations of algorithmic matching,
each of which brings with
it a different risk of making an incorrect match. These variations can be
invoked by either the \fB-e\fR or the \fB-g\fR option in the \fBbulkload\fR
command line.
.FS
The \fB-g\fR option was only made available in BLT in June 1994. Users of
earlier versions of the software should obtain the latest version from UCL.
.FE
The \fB-e\fR option will match data from different sources only if all of the
following conditions apply:
.IP a.
Both surnames are the same
.IP b.
Any personal titles, if any, within the RDNs are consistent e.g. Mr.
and Dr. will match but Mr. and Miss will not.
.IP c.
Both entries have the same number of forenames and/or initials
.IP d.
Both sets of forenames and/or initials match
.LP
.PP
So Paul F Barker will match with Dr P F Barker but not P Barker.
.PP
The \fB-g\fR option will first apply the matching algorithm above.
If that finds an
unambiguous match, it accepts that match. Otherwise, it tries to match replacing
last two conditions with:
.IP c\(fm.
Both sets of forenames and/or initials match to the end of the shorter one
.PP
So Paul F Barker will match with Dr P F E Barker and P F E Barker.
.PP
The \fB-g\fR option is useful where one data source contains fewer initials or
forenames than other sources. Obviously there is a greater chance of making an
incorrect match with the \fB-g\fR option than the \fB-e\fR option.
To allow you to monitor
its decisions, the \fB-g\fR option reports any matches that it made but which
would have been rejected by the \fB-e\fR option, to the .done file.
.NH 1
Strategy for Loading Leaf Entries
.PP
.nh
Having loaded the dataSources and the OUs, the following steps should be taken
in loading the data from each dataSource,
starting with the most authoritative ones.
.IP 1.
Prepare an `empty' \fIdmformat\fR file which contains only the \fBsource\fR
line, e.g. for data from dataSource adminStaff the file should contain only
the line:
.RS
.IP
.ft C
source=adminStaff
.ft
.RE
.IP 2.
Use \fBnewdmdiff\fR to make a file of the updates for \fBbulkload\fR,
for example, by: 
.RS
.IP
newdmdiff \fIoldfile newfile\fR >staffdata
.RE
.LP
.PP
This will produce the update file in staffdata.
(Of course, on this first run, it would be sufficient to skip Steps 1 and 2,
and run the \fIdmformat\fR file through \fBbulkload\fR.)
.KS
.IP 3.
Load new entries and update non-interactively using exact
matching. This can be invoked using:
.RS
.IP
bulkload -o n
.RE
No potentially ambiguous updates to existing entries will be performed.
However, these cases \fIwill\fR be reported to the .PROBLEM file. For example,
using:
.RS
.IP
bulkload staffdata -o n
.RE
will report all potentially ambiguous matches to staffdata.PROBLEM.
.KE
.IP 4.
By this stage all new entries will have been added.
Consequently, the lines in the file which add attributes
like objectClass, need not be included for the later runs.
Attempting to add attributes which are already in the entries
will not cause problems with \fBbulkload\fR's updates. However
they will be reported as warnings (probably  to the .noAction file).
Editing out these lines from the .PROBLEM file from Step 3
will reduce the size of the
report files produced in later runs.
.IP 5.
As mentioned earlier, the .PROBLEM file can be input to another run of
\fBbulkload\fR. However, this will produce journal files with suffixes as well.
So we will get files with names like staffdata.PROBLEM.PROBLEM.
Therefore, link the .PROBLEM file from step 4 to another name.
For example, use:
.RS
.IP
ln staffdata.PROBLEM st.P
.RE
.IP 6.
Process updates from the problem file non-interactively using algorithmic
matching. This can be invoked by:
.RS
.IP
bulkload st.P -o n -e
.RE
or
.RS
.IP
bulkload st.P -o n -g
.PP
This run sends any matches that it thinks are ambiguous to a .PROBLEM file,
in this case st.P.PROBLEM.
.RE
.IP 7.
By now you should have succeeded in adding all but a small fraction of the
entries from \fIthis\fR source.
Most of these cases will have been rejected as `ambiguous' by
\fBbulkload\fR.
Try to resolve these cases before resubmitting them.
.sp
Look at st.P.PROBLEM to see these entries. They can only be resolved by the
your knowledge, outside that which you have already provided in the directory.
For each entry, you need to get as much information as possible to help
resolve it.
(If you are working in a windowing environment,
you will probably find it useful to use one window to sort things out while
\fBbulkload\fR runs in another window.)
In general, your approaches may include:
.RS
.IP a.
Using your local knowledge of the people involved, e.g. knowing that J M Fraser
is always known as Malcolm may resolve a potential ambiguity.
.RE
.RS
.IP b.
Using \fBfdish\fR, move to the OU where the ambiguous entry is to be placed.
Then do as search on the surname of the entry.  Display the contents of all
The entries returned. These entries may help you work out who is who.
.RE
.RS
.IP c.
Looking at information from other sources e.g. old internal telephone books
may be differently orgainsed from current lists and so provide other evidence.
.RE
.RS
.IP d.
Phoning the department concerned. This may be the only way to find out whether
two people with similar entries are the same person or are distinct, e.g.
in my department we \fIdo\fR have two Christopher Johnsons.
.RE
.RS
.IP e.
Entering the data into the directory and waiting. People who are misclassified
may eventually contact you. This will allow you to correct the entry in a later
run.
.RE
.LP
.IP 8.
Having resolved as much as possible,
these entries can only be added
by an interactive run, which is invoked by:
.RS
.IP
bulkload st.P.PROBLEM
.RE
.PP
When you have finished adding the data from one source, you can repeat these
steps for other data sources.
.NH 1
Incorporating Changes to the Directory with Newdmdiff
.PP
.nh
After you have loaded the directory using \fBbulkload\fR,
you can use \fBnewdmdiff\fR to 
help with making regular updates e.g. when phone numbers change.
\fBNewdmdiff\fR will produce files which delete old attributes and add
their replacements. It will also delete entries which are not in the new file
and add in new entries.
.PP
.KS
An example of the output produced by \fBnewdmdiff\fR is:
.ft C
.B
.TS
box;
l.
newdmformat [1.1]

source=emailData

# Notation:-
# $XXX indicates that XXX is an       RDN       that is to be ADDED
# !XXX indicates that XXX is an       RDN       that is to be DELETED
# +XXX indicates that XXX is an    ATTRIBUTE    that is to be ADDED
# -XXX indicates that XXX is an    ATTRIBUTE    that is to be DELETED
# >XXX indicates that XXX is an ATTRIBUTE VALUE that is to be ADDED
# <XXX indicates that XXX is an ATTRIBUTE VALUE that is to be DELETED


rootedAt=c=GB@o=Brunel University@ou=Administration


cn= A McLaughlin
>userid= acsram
<userid= acsraam

!cn= A N Fisher

$cn= A N Fischer
>cn= Arthur N Fischer
+sn= Fisher
+rfc822Mailbox= Arthur.Fischer@brunel.ac.uk
+userid= acsranf
+objectClass= top 
>objectClass= person 
>objectClass= organizationalPerson 
>objectClass= pilotObject 
>objectClass= newPilotPerson 
>objectClass= quipuObject


rootedAt=c=GB@o=Brunel University@ou=Biology and Biochemistry



rootedAt=c=GB@o=Brunel University@ou=Brunel Institute of Bio-Engineering

# etc. . .
.TE
.KE
.ft
.R
.KS
.PP
In general, to incorporate updates, you should:
.IP 1.
Obtain the latest version of data for dataSource.
.KE
.IP 2.
Convert this new data to \fIdmformat\fR using a method
similar to the \fBperl\fR scripts in the appendixes.
.sp
You must use
the last version of each of the \fIdmformat\fR files
which was successfully applied to the directory,
as the \fIoldfile\fR for \fBnewdmdiff\fR.
.IP 3.
Load the leaf entries for this new data to using 
the operations of the previous Section, starting from Step 2.
Hopefully, this process can be essentially automatic because new
`ambiguous' data should only occur rarely.
.PP
Do not worry if this phase fails due to external problems e.g. a machine crash.
BLT is resilient enough to allow you to repeat the loading phase from the start,
so long as you have used the corrct \fIoldfile\fR in the previous Step.
All that will happen is that a large number of lines will be written to
the .noAction file when you try to repeat the parts of the update
which have already succeeded.
This will not occur on the run after an update has run properly to completion.
.NH 1
Converting a Currently Operating Directory to BLT
.PP
Assume that you want to move, as much as possible, to BLT
from a directory which is not currently using it.
You should never expect to maintain all of the the directory
with BLT alone, because you must expect that the owners of an entry 
or local directory administrators may
update at least a subset of the entries' attributes using tools other
than BLT e.g. \fBdish\fR.
These changes will not be discovered by BLT.
The list of attribute values that are present in the directory but not
in the input data sources may be considerable, but the values may be
perfectly valid (without you having any knowledge of
which are correct and which are not).
.PP
This `extra' data can consist of four classes:
.IP a.
Entries which are in the directory prior to the introduction of BLT
and which remain outside BLT.
A way to identify these entries is discussed in Section 14.
.IP b.
Entries which are added to the directory by other DUAs after
the introduction of BLT and which remain `undiscovered' by BLT.
.IP c.
Attributes supplied by DUAs other than bLT
for entries that BLT supplies other attributes
e.g. BLT does not supply facsimileTelephoneNumber but does supply
telephoneNumber.
BLT `discovers' these entries but only the attributes it supplies.
.IP d.
Attributes for which BLT provides \fIsome\fR values but not all.
Once \fBbulkload\fR has added one value, it will treat \fIall\fR
values of \fIthese\fR attributes as its own.
.sp
.PP
In addition to the facilities described earlier, \fBbulkload\fR
provides a set of bulk management functions.
They can also be used to help identify entries created prior to BLT,
although they were designed to maintain consistency between the
directory and the various dataSources.
.PP
Before using the bulk management functions, it is important to ensure
that you have ACL read access to all of the attributes in the entries.
If you do not have complete access, bulk management will report
a large number of spurious differences.
.PP
The bulk management functionality is accessed by:
.RS
.IP
bulkload -m
.RE
.LP
.PP
It presents you with a menu of options. These
include the ability to dump a subdirectory or directories
to a file in \fIdmformat\fR (Function 3). They also allow you to
determine whether changes in the dataSources that are being used by BLT
have been reflected in the directory
(Functions 4 and 5). Bulk management can report on the entries which have
\fInot\fR been affected by a series of BLT updates (Function 6).
.sp
.PP
Before you start the conversion process,
you should consider whether it is worth building the directory
from scratch. This can be a viable option if
most of the current contents of the directory data is out of date,
or is readily available in BLT format.
It can also be worthwhile if you want to perform a significant reorganisation
of the directory structure.
Function 3 of the management functionality can help in producing
a set of directory input files by dumping out the previous entries.
You could also use the EDB files as a source for the BLT data.
.PP
Alternatively, if you decide to merge your current directory
with the data provided by BLT,
you will need the data from a number of `new' dataSources.
However these dataSources are unlikely to
provide all the information which is currently in the directory.
The full process is:
.IP 1.
Add dataSource infrastructure entries for the `new' dataSources to the current
directory as discussed in Section 7 above.
.IP 2.
For each dataSource do the following:
.RS
.IP a.
Apply new updates of data from dataSource to the directory using the methods
of Section 11 to resolve ambiguities.
.sp
.KS
This adds any new data from this data source. It
leaves alone `changes' which are the same as current values. It will also
not affect attributes which are not mentioned,
and attribute values which are extra to those in `new'
source, e.g. `extra' phone numbers which the `new' phone list
may not know about.
.sp
It also adds an algoName attribute for this dataSource to
each entry it `touches'.
.KE
.IP b.
Compare the values of the attributes added by the dataSource with the
values in the dataSource file. Any `extra' values are pre-existing
values which should be checked to see whether they are still appropriate.
You will need to use \fBdish\fR or another DUA to do this because the
bulk managament functions only attempt to monitor attribute values
that BLT has added
(This approach is taken by BLT because it recognises that
there will be attributes
added by other DUAs which it knows nothing about).
.RE
.NH
Entries Outside BLT
.PP
In the previous Section, it was noted that
a number of entries in the directory from prior to the introduction of BLT,
may never be discovered by it.
As part of the transition to a BLT managed directory, BLT \fIcan\fR be
used to identify the entries from the old directory that it does not
normally reach.  Note that, as explained below, the method described
here will not work as a continuing method for identifying entries `outside'
BLT.
.PP
We assume that every subdirectory has the \fBbulkDataEntryDetails\fR
attribute in it.
This is guaranteed if the infrastructure has been set up or updated using
\fBbulkload\fR with a file similar to that produced by
\fBmake-ou-load-file\fR.
\fBAfter\fR you have completed the updates from your BLT dataSources,
perform the following:
.IP 1.
Run bulk management Function 4.
This will set the \fBvalidationCheckAt\fR attributes in all the entries
within the sub-tree to the time of this run.
.IP 2.
Run each of the latest dataSource files
through bulk management Function 5, one by one (Function 5 presents a menu
requesting each file name in turn).
Function 5 looks for the algoName attribute in each entry in the directory
corresponding to the data in its input file.
It then checks that all the updates from these data sources have been applied
to the directory.
It also updates the \fBvalidationCheckAt\fR
attribute to the time when the comparison occurred. 
So this step will touch all entries which have at least one attribute
in the dataSources.
(Function 5 produces a \fIdmformat\fR file for the attributes which are in the
input files but not in the directory.
Because we have previously applied all these updates, we would
expect this file to be empty.)
.IP 3.
Run bulk management Function 6. This checks for all those entries
whose \fBvalidationCheckAt\fR is still set to the time value applied
in Step 1,
i.e. those entries that have not be compared against by any of the 
data sources and were thus added from unknown sources.
.PP
All of the entries found by Function 6 will also have the \fBlastModifiedTime\fR attribute older
than the data of the first run through management Function 5.
So, these entries were in the directory before BLT was introduced.
They may be still valid but we don't know.
They may have come from sources we do not have within BLT's control
e.g. \fBdish\fR input for external people like visiting login accounts, direct
dial phone lines which are not reported in the internal phone books, etc.
They may be truly expired entries.
.PP
We can keep this list of `old' entries.
Any subsequent entries from outside BLT
will get a \fBlastModifiedTime\fR
after the run through Function 5.
So we can identify the entries which are outside BLT but are still current.
.PP
Occasionally we can use \fBdish\fR, or another DUA, to report on
all the `old' entries.
Any changes to \fBlastModifiedTime\fR
for these ancient but not forgotten entries
will indicate that they are still required.
.PP
As we improve the number of data sources and the supply of
data to BLT,
many of the `old' entries will become known to BLT from other sources.
For other entries we will get specific
information telling us that they have expired. So they can be deleted.
So long as there are not any major amendments made to the directory
which change the \fBlastModifiedTime\fR of
the `old' directory entries
but do not change them otherwise,
over time
the number of entries which are unknown to BLT should
become less significant.
.NH 1
Conclusions
.PP
.nh
This paper has provided an outline of how to use BLT. It has shown
how to prepare data for loading into the directory and described the steps
in performing the loading.
It has explained how \fBbulkload\fR matches entries from
different data sources.
It has also suggested a method for converting from a
directory which is not using BLT to one supported by BLT.
The paper is supported by a number of \fBperl\fR scripts
which provide guidance in preprocessing the data into \fBbulkload\fR's
input format.
.PP
The author would like to acknowledge the assistance given by Andrew Findlay
in the preparation of this paper.
He also appreciates the comments given on earlier drafts by Paul Barker
and Kevin McCarthy.
.SH
References:

.IP [1]
Barker, P. and McCarthy K., "Technical Details and User Guide for the
Bulk Loading Tools Newdmdiff and Bulkload", Department of Computer
Science, University College London, April 1993.

.IP [2]
ISO/IEC, "The Directory --- Overview of concepts, models and services",
Recommendation X.500, ISO/IEC 9594-1, 1993.

.IP [3]
Barker, P., Kille, S. and Lenggenhager, T.  "Naming and Structuring
Guidelines for X.500 Directory Pilots", RFC 1617, May 1994.

.IP [4]
Michaelson, G. and  Prior, M., "Naming Guidelines for the AARNet
X.500 Directory Service", RFC 1562, AEN-001, December 1993.

.IP [5]
Barker, P., "Preparing data for inclusion in an X.500
Directory", Research Note RN/92/41, Department of Computer
Science, University College London, May 1992.

.IP [6]
ISODE Consortium, "Administrator's Guide Directory Services", 
ISODE Version 1.0 Volume 5, ISODE Consortium, 1993.
.LP
