#include <isode/quipu/util.h>
#include <isode/quipu/add.h>
#include <isode/quipu/remove.h>
#include <isode/quipu/modifyrdn.h>
#include <isode/quipu/dua.h>

#include "util.h"
#include "bulkloaderrors.h"
#include "referral.h"


int
AddEntry( entryDN, as, errorMsgcpp, errorClassp, errorTypep )
DN	entryDN ;
Attr_Sequence as ;
char **	errorMsgcpp ;
int  *	errorClassp, * errorTypep ;
{
int 	ret_val = NOTOK, loopCount = 0 ;
static 	struct  ds_addentry_arg aarg ;
static 	struct  DSError aerror ;
static 	CommonArgs ada_common_args = default_common_args ;
#ifdef BLT_DEBUG
char *	dncp ;
#endif
	if (rebind() != OK)
	    return NOTOK;

	aarg.ada_common = ada_common_args ; /* struct copy */
	aarg.ada_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	aarg.ada_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;

	/* Define the entry to be removed.. */
#ifdef BLT_DEBUG
fprintf( stderr, "\n>>Add entry 1 dn %s\n",(dncp= dn2pstr( entryDN )) ) ;
(void)free( dncp ) ;
#endif
	aarg.ada_object = entryDN ;

	aarg.ada_entry = as ;

	
	/* Request the specifed attribute modifications */

	do {

	    startUnbindTimer() ;
	    ret_val = de_addentry( &aarg, &aerror );
	    stopUnbindTimer() ;

	} while( loopCount++ < 1 && referralToBeProcessed( ret_val, &aerror ) == TRUE ) ;


	if (ret_val != OK)
	{
	    (void)errorClassAndType( &aerror, 0, errorClassp, errorTypep, NULLAttrT, NULLAttrV ) ;

	    *errorMsgcpp = error2pstr( &aerror ) ; /* ds_error_free() called by ds_error() */

#ifdef BLT_DEBUG
	    fprintf( stderr, "%s\n", *errorMsgcpp ) ;
#endif
    	    return NOTOK;
	}
	else *errorMsgcpp = NULLCP ;

	return( ret_val );
}


int
RemoveEntry( entryDN, errorMsgcpp )
DN	entryDN ;
char **	errorMsgcpp ;
{
int 	ret_val = NOTOK, loopCount = 0 ;
static 	struct  ds_removeentry_arg rarg ;
static 	struct  DSError rerror ;
static 	CommonArgs rea_common_args = default_common_args ;

	if (rebind() != OK)
	    return NOTOK;

	rarg.rma_common = rea_common_args ; /* struct copy */
	rarg.rma_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	rarg.rma_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;

	/* Define the entry to be removed.. */

	rarg.rma_object = entryDN ;

	
	/* Request the specifed attribute modifications */
	do {

	    startUnbindTimer() ;
	    ret_val = de_removeentry( &rarg, &rerror );
	    stopUnbindTimer() ;

	} while( loopCount++ < 1 && referralToBeProcessed( ret_val, &rerror ) == TRUE ) ;


	if (ret_val != OK)
	{    
	    *errorMsgcpp = error2pstr( &rerror ) ; /* ds_error_free() called by ds_error() */
#ifdef BLT_DEBUG
	    fprintf( stderr, "%s\n", *errorMsgcpp ) ;
#endif
    	    return NOTOK;
	}
	else *errorMsgcpp = NULLCP ;
	
	return( ret_val );
}


int
ModifyRdn( entryDN, newRDN, removeOldRDNAttrs, errorMsgcpp )
DN	entryDN ;
RDN	newRDN ;
char	removeOldRDNAttrs ;
char **	errorMsgcpp ;
{
int 	ret_val = NOTOK, loopCount = 0 ;
static 	struct  ds_modifyrdn_arg marg ;
static 	struct  DSError merror ;
static 	CommonArgs mra_common_args = default_common_args ;

	if (rebind() != OK)
	    return NOTOK;

	marg.mra_common = mra_common_args ; /* struct copy */
	marg.mra_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	marg.mra_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;

	/* Define the entry to be removed.. */
#ifdef BLT_DEBUG
fprintf( stderr, "\n>>Modify rdn entry 1 dn %s\n", dn2pstr( entryDN ) ) ;
#endif
	marg.mra_object = entryDN ;

	marg.mra_newrdn = newRDN ;

	marg.deleterdn  = removeOldRDNAttrs ;


	/* Request the specifed attribute modifications */

	do {

	    startUnbindTimer() ;
	    ret_val = de_modifyrdn( &marg, &merror );
	    stopUnbindTimer() ;

	} while( loopCount++ < 1 && referralToBeProcessed( ret_val, &merror ) == TRUE ) ;



	if (ret_val != OK)
	{
	    *errorMsgcpp = error2pstr( &merror ) ; /* ds_error_free() called by ds_error() */

#ifdef BLT_DEBUG
	    fprintf( stderr, "%s\n", *errorMsgcpp ) ;
#endif
    	    return NOTOK;
	}
	else *errorMsgcpp = NULLCP ;

	return( ret_val );
}
