#include <isode/quipu/name.h>
#include <isode/quipu/ds_search.h>
#include <isode/quipu/read.h>
#include <isode/quipu/dua.h>

#include "util.h"
#include "namelist.h"
#include "filter.h"
#include "bulkloadglobal.h"
#include "referral.h"

struct ds_search_arg *
fillMostSearchArgs( parentDN, searchdepth, attrsToBeReturnedp, allAttributes )
DN 	parentDN;
int 	searchdepth;
struct 	namelist * attrsToBeReturnedp ;
int	allAttributes ;
{
static struct ds_search_arg sarg;
Attr_Sequence * atl;
AttributeType at;
struct namelist * attrListp;
static CommonArgs sca = default_common_args;

        sarg.sra_common = sca; /* struct copy */
	sarg.sra_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	sarg.sra_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;

	sarg.sra_common.ca_servicecontrol.svc_options  |= SVC_OPT_DONTUSECOPY ;
	sarg.sra_common.ca_servicecontrol.svc_options  |= SVC_OPT_DONTDEREFERENCEALIAS ;
	
	sarg.sra_subset 	= searchdepth;
	sarg.sra_baseobject 	= dn_cpy( parentDN ) ;
	sarg.sra_searchaliases 	= FALSE;
	
	/* specify attributes of interest */
	if( attrsToBeReturnedp != NULLLIST )
	{
	    sarg.sra_eis.eis_allattributes = FALSE;

	    atl = &(sarg.sra_eis.eis_select);

            for( attrListp = attrsToBeReturnedp ; attrListp != NULLLIST; attrListp = attrListp->next)
	    {
#ifdef BLT_DEBUG
fprintf( stderr, "To be returned %s\n", attrListp -> name ) ;	    
#endif
            	if ((at = str2AttrT( attrListp -> name)) == NULLAttrT)
		    continue;

	    	*atl = as_comp_alloc();
	    	(*atl)->attr_type = at;
	    	(*atl)->attr_value = NULLAV;
	    	atl = &(*atl)->attr_link;
	    }

	    *atl = NULLATTR ;
	}
	else
	{
	    sarg.sra_eis.eis_allattributes = allAttributes ;
	    sarg.sra_eis.eis_select = NULLATTR ;
	}
	
	sarg.sra_eis.eis_infotypes = EIS_ATTRIBUTESANDVALUES;

	return ( &sarg );
}



void
freeSearchArgs( sargp )
struct ds_search_arg * sargp ;
{
	dn_free( sargp -> sra_baseobject);
	/* Note: filter_free( sargp -> sra_filter ) ; is called in RetrieveAndListAttrs() */
	as_free( sargp -> sra_eis.eis_select);
}


/* 
 * Carry out the search and produce a list of results
 */

int
RetrieveAndListAttrs( searchListpp, sargp, sresultp, serrorp )
struct	namelist ** searchListpp ;
struct	ds_search_arg 	 * sargp ;
struct	ds_search_result * sresultp ;
struct	DSError 	 * serrorp ;
{
int	ret_val = SEARCHOK, res, loopCount=0 ;
entrystruct * entryp ;
struct	namelist ** currListpp = searchListpp ;

        if (rebind() != OK)
{
#ifdef BLT_DEBUG
fprintf( stderr, "Bind problem \n" ) ;
#endif
	    return SEARCHFAILBINDPROBLEM ;
}
#ifdef BLT_DEBUG
else fprintf( stderr, "Bind OK\n" ) ;
#endif

	do {

	    startUnbindTimer() ;
	    res = de_search( sargp, serrorp, sresultp ) ;
	    stopUnbindTimer() ;

	} while( loopCount++ < 1 && referralToBeProcessed( res, serrorp ) == TRUE ) ;
	
        if (res == DSE_INTR_ABANDONED)
	    if (serrorp->dse_type == DSE_ABANDONED)
		ret_val = SEARCHABANDONED ;
	    else if (serrorp->dse_type == DSE_SECURITYERROR)
		ret_val = SEARCHFAILSECPROBLEM ;

	filter_free( sargp -> sra_filter);

	if (res != OK)
	{
 	    /* Note: the DSError has not been freed */
	    return SEARCHNOTOK;
	}

	 /* Ensure one element in list of search results.. */
	correlate_search_results( sresultp );


        for( entryp = sresultp -> CSR_entries; 
		entryp != NULLENTRYINFO; entryp = entryp->ent_next )
	{	
	    *currListpp 	  = list_alloc();
	    
	    (*currListpp)->name   = dn2pstr( entryp->ent_dn );

	    (*currListpp)->entryDN = dn_cpy( entryp->ent_dn );
	    
	    (*currListpp)->ats    = as_cpy( entryp->ent_attr );
            
	    currListpp 	  	  = &(*currListpp)->next;
	}

        *currListpp = NULLLIST;


	/* Return storage.. */
	
	entryinfo_free( sresultp -> CSR_entries, 0 );
	dn_free( sresultp -> CSR_object );
	crefs_free( sresultp -> CSR_cr);


	return( ret_val ) ; /* SEARCHNOTOK, SEARCHOK,... [bulkloadglobal.h] */
}




int
readEntry( entryDN, plistpp, attrsToBeReturnedp, allAttributes )
DN	entryDN ;
struct 	namelist ** plistpp;
struct 	namelist  * attrsToBeReturnedp ;
int	allAttributes ;
{
static 	struct ds_read_arg rarg;
static 	struct ds_read_result rresult;
static 	struct DSError rerror;
static 	CommonArgs sca = default_common_args;
Attr_Sequence * atl;
AttributeType at;
struct namelist * x;
int 	retval = OK, loopCount=0 ;

	if (rebind() != OK)
		return NOTOK;

	rarg.rda_common =sca; /* struct copy */
	rarg.rda_common.ca_servicecontrol.svc_timelimit = SVC_NOTIMELIMIT;
	rarg.rda_common.ca_servicecontrol.svc_sizelimit = SVC_NOSIZELIMIT;

	rarg.rda_common.ca_servicecontrol.svc_options  |= SVC_OPT_DONTUSECOPY ;
	rarg.rda_common.ca_servicecontrol.svc_options  |= SVC_OPT_DONTDEREFERENCEALIAS ;

	rarg.rda_object = dn_cpy( entryDN ) ;

	/* specify attributes of interest */

	if( attrsToBeReturnedp != NULLLIST )
	{
	    rarg.rda_eis.eis_allattributes = FALSE ;
	
	    atl = &(rarg.rda_eis.eis_select);

	    for( x = attrsToBeReturnedp ; x != NULLLIST; x = x->next)
	    {
		if ((at = str2AttrT(x->name)) == NULLAttrT)
			continue;

		*atl = as_comp_alloc();
		(*atl)->attr_type = at;
		(*atl)->attr_value = NULLAV;
		atl = &(*atl)->attr_link;
	    }
	
	    *atl = NULLATTR;
	}
	else
	{
	    rarg.rda_eis.eis_allattributes = allAttributes ;
	    rarg.rda_eis.eis_select = NULLATTR ;
	}
	
	rarg.rda_eis.eis_infotypes = EIS_ATTRIBUTESANDVALUES;

	do {

	    startUnbindTimer() ;
	    retval = de_read(&rarg, &rerror, &rresult);
	    stopUnbindTimer() ;

	} while( loopCount++ < 1 && referralToBeProcessed( retval, &rerror ) == TRUE ) ;

	if (retval != OK)
	{
#ifdef BLT_DEBUG
	    fprintf( stderr, "Read error %s\n", error2pstr( &rerror ) ) ;
#else
	    ds_error_free( &rerror ) ;
#endif
	    retval = NOTOK;
	}
	else
	{
	    *plistpp = list_alloc();
	    (*plistpp)->name    = dn2pstr(rresult.rdr_entry.ent_dn);
	    (*plistpp)->entryDN = rresult.rdr_entry.ent_dn ;
	    rresult.rdr_entry.ent_dn = NULLDN ;
	    
	    (*plistpp)->ats  = rresult.rdr_entry.ent_attr ;
	    rresult.rdr_entry.ent_attr = NULLATTR ;
	    
	    (*plistpp)->next = NULLLIST;
	}
	
	dn_free( rarg.rda_object ) ;
	as_free( rarg.rda_eis.eis_select );
	
	return retval;
}
