#include <isode/quipu/name.h>


#include	"bulkloadattrnames.h"
#include	"list.h"
#include	"namelist.h"
#include	"subdirectory.h"
#include	"util.h"


void
initSubDirectoryList( subDirListp )
struct  subDirListDetails * subDirListp ;
{
	subDirListp->subDirListRDN	= NULLRDN ;

	subDirListp->subDirListUsed	= SUBDIRLISTNOTUSED ;
	subDirListp->rootedAtDN		= NULLDN ;
	subDirListp->bulkSubTreeDN	= NULLDN ;
	subDirListp->bulkSubTreeDNextensions = ((DN *)NULL) ;

	return ;
}

/*
 * Ensure that the results are all from the Sub tree directories that has been specified..
 */

void
filterSubTreeNamelist( rootedAtSubDirListDetailsp, retrievedListpp )
struct  subDirListDetails * rootedAtSubDirListDetailsp ;
struct 	namelist ** retrievedListpp ;
{
struct 	namelist * entryListp, * removeEntryListp, * prevEntryListp = NULLLIST ;
DN	entryDnComp = NULLDN, buffEntryDnComp, subTreeDnComp ;
DN	subTreeDNext ;
int	matchOK = TRUE, DNindex ;

	if( rootedAtSubDirListDetailsp -> subDirListUsed == SUBDIRLISTNOTUSED )
	    return ;

	for( entryListp = *retrievedListpp ; entryListp != NULLLIST ; )
	{
	    matchOK = TRUE ;
	    
	    if( rootedAtSubDirListDetailsp -> subDirListUsed == SUBDIRLISTALLSUBORDSUSED )
	    {
	    	/* Check that the entryDN has only 2 DN components (one for the extension 
	    	   e.g. OU and one for the entry RDN) more than the subTree parent DN..   */

	    	if( countDNcomps( entryListp->entryDN ) != (countDNcomps( rootedAtSubDirListDetailsp->bulkSubTreeDN ) + 2) )
		    matchOK = FALSE ;
	    }
	    else if( rootedAtSubDirListDetailsp -> subDirListUsed == SUBDIRLISTUSED )
	    {
		/* First match against the components of the sub directory list parent DN.. */
	    	matchOK = TRUE ;

		/* The retrieved entry DN should  have LESS DN components than the sub directory parent.. */
		if( entryListp->entryDN == NULLDN || rootedAtSubDirListDetailsp->bulkSubTreeDN == NULLDN ||
		    countDNcomps( entryListp->entryDN ) <= countDNcomps( rootedAtSubDirListDetailsp->bulkSubTreeDN ) )
		{
		    matchOK = FALSE ;
		}
		else
		{
		    for( entryDnComp = entryListp->entryDN, subTreeDnComp = rootedAtSubDirListDetailsp->bulkSubTreeDN ; 
		     	 entryDnComp != NULLDN && subTreeDnComp != NULLDN ;
		     	 entryDnComp = entryDnComp->dn_parent, subTreeDnComp = subTreeDnComp->dn_parent )
		    {	 /* ..Note: entryDnComp may be used after this loop */

		    	if( rdn_cmp( entryDnComp->dn_rdn, subTreeDnComp->dn_rdn ) != 0 )
		    	{
		    	    matchOK = FALSE ;
			    break ;
		    	}
		    }
		}

		if( matchOK == TRUE )
		{
		    matchOK = FALSE ;

		    /* Now check that the next DN components are one of those listed in the list of "bulkSubTreeDNextension"s */
		    if( entryDnComp != NULLDN )
		    {
			buffEntryDnComp = entryDnComp ;
		    
		    	/* For each DN extention.. */
		    	for( DNindex = 0, subTreeDNext = *(rootedAtSubDirListDetailsp->bulkSubTreeDNextensions + DNindex) ;
			     matchOK == FALSE && subTreeDNext != NULLDN ;
			     ++DNindex, subTreeDNext = *(rootedAtSubDirListDetailsp->bulkSubTreeDNextensions + DNindex) )
			{
			    /* For each DN component within a DN extension.. */
			    for( entryDnComp =  buffEntryDnComp, subTreeDnComp = subTreeDNext ;
			         entryDnComp != NULLDN && subTreeDnComp != NULLDN ; 
				 entryDnComp =  entryDnComp->dn_parent, subTreeDnComp = subTreeDnComp -> dn_parent )
			    {
				if( rdn_cmp( entryDnComp->dn_rdn, subTreeDnComp->dn_rdn ) != 0 )
			    	{
				    break ; /* Did not match against this extension component */
				}
			    }

			    if( subTreeDnComp == NULLDN )
			    {
			    	 /* All the DN components in the extension have been matched */
				 
				/* One final check to ensure that there is only one further entry DN component */
				if( entryDnComp != NULLDN && entryDnComp -> dn_parent == NULLDN )
				    matchOK = TRUE ;
			    }
			}
		    }
		}
	    }

	    if( matchOK == FALSE )
	    {
#ifdef BLT_DEBUG
fprintf( stderr, "Filtering out %s\n", entryListp -> name ) ;
#endif
		/* This entry's DN is not within a specified sub directory */

		removeEntryListp = entryListp ;

		/* Check if this is the base entry.. */
		if( entryListp == *retrievedListpp )
		    entryListp = (*retrievedListpp) = (*retrievedListpp) -> next ;
		else
		    entryListp = entryListp -> next ;

		/* Link around the removed node.. */
		if( prevEntryListp != NULLLIST )
		    prevEntryListp -> next = entryListp ;

		/* Remove the entry node.. */
		removeEntryListp->next = NULLLIST ;
		freeListpp( &removeEntryListp ) ;
	    }
	    else 
	    {	
#ifdef BLT_DEBUG
fprintf( stderr, "Filtering IN %s\n", entryListp -> name ) ;
#endif
	    	prevEntryListp = entryListp ;
	    	entryListp = entryListp -> next ;
	    }
	}
	
	return ;
}

/*
 * After the particular sub-directory is determined further
 * filter the retrieved list so that all entries belong to it
 */

void
filterSubTreeNamelistForSingleSubDirectory( subDirListParentDN, retrievedListpp )
DN	subDirListParentDN ;
struct 	namelist ** retrievedListpp ;
{
struct 	namelist * entryListp, * removeEntryListp, * prevEntryListp = NULLLIST ;
DN	entryDnComp, subDirListParentDnComp ;
int	matchOK = TRUE ;

	for( entryListp = *retrievedListpp ; entryListp != NULLLIST ; )
	{
	    matchOK = TRUE ;

	    for( entryDnComp = entryListp->entryDN, subDirListParentDnComp = subDirListParentDN ; 
		 entryDnComp != NULLDN && subDirListParentDnComp != NULLDN ;
		 entryDnComp = entryDnComp->dn_parent, subDirListParentDnComp = subDirListParentDnComp->dn_parent )
	    {
		if( rdn_cmp( entryDnComp->dn_rdn, subDirListParentDnComp->dn_rdn ) != 0 )
		{
		    matchOK = FALSE ;
		    break ;
		}
	    }

	    if( matchOK == TRUE )
	    {
		/* All the DN components in the extension have been matched */
				 
		/* One final check to ensure that there is only one further entry DN component */
		if( subDirListParentDnComp != NULLDN || entryDnComp == NULLDN || entryDnComp -> dn_parent != NULLDN )
		    matchOK = FALSE ;
	    }

	    if( matchOK == FALSE )
	    {
#ifdef BLT_DEBUG
fprintf( stderr, "Filtering out %s\n", entryListp -> name ) ;
#endif
		/* This entry's DN is not within a specified sub directory */

		removeEntryListp = entryListp ;

		/* Check if this is the base entry.. */
		if( entryListp == *retrievedListpp )
		    entryListp = (*retrievedListpp) = (*retrievedListpp) -> next ;
		else
		    entryListp = entryListp -> next ;

		/* Link around the removed node.. */
		if( prevEntryListp != NULLLIST )
		    prevEntryListp -> next = entryListp ;

		/* Remove the entry node.. */
		removeEntryListp->next = NULLLIST ;
		freeListpp( &removeEntryListp ) ;
	    }
	    else 
	    {	
#ifdef BLT_DEBUG
fprintf( stderr, "Filtering IN %s\n", entryListp -> name ) ;
#endif
	    	prevEntryListp = entryListp ;
	    	entryListp = entryListp -> next ;
	    }
	}
	
	return ;
}

/*
 * Check if the last DN component within a "rootedAt" DN contains a "bulkSubDirList" attribute.
 * If it does, check if the list is already held. Before retrieving the "bulkSubTreeDN" and
 * "bulkSubTreeDNextension" attributes from the list's Directory entry check that the list name is not 
 * the special "All" subordinates case.
 *
 * When the "bulkSubTreeDNextension" DNs are retrieved they are stored in an array for return.
 *
 * NOTE: the subordinate DNs for the "All" subordinates special case are not set up here 
 * as they are retrieved only when necessary e.g. within userInputEntryOrParentDNresolution().
 *
 */

int
checkForAndExpandSubDirList( rootedAtSubDirListDetailsp, rootedAtDN, errorDiagnosticscp )
struct  subDirListDetails * rootedAtSubDirListDetailsp ;
DN	rootedAtDN ;
char *	errorDiagnosticscp ; /* Pointer to SMALLBUFLEN buffer */
{
int	retVal = OK, DNindex=0, DNcount=0 ;
char *	DNcp = NULLCP, * subDirListNamecp = NULLCP ;
DN  *	baseDNextList ;
RDN	lastRootedAtRDN, subDirListRDN ;
DN	nextToLastRootedAtDN ;
struct	namelist * attrsToBeReturnedp = NULLLIST, * readSubDirListp = NULLLIST ;
Attr_Sequence	as  ;
AV_Sequence	avs ;
static	AttributeType	bulkSubDirListat = NULLAttrT ;


	if( bulkSubDirListat == NULLAttrT )
	    bulkSubDirListat =  str2AttrT( BL_BULK_SUB_DIR_LIST ) ;
	    
	/* Check if the last RDN of the "rootedAt" DN contains a sub directory list attribute.. */
	if( (lastRootedAtRDN = findLastRDN(rootedAtDN)) == NULLRDN ||
	    (subDirListRDN   = findAttrTypeRdnComp( bulkSubDirListat, lastRootedAtRDN )) == NULLRDN )
	{
	    /* .. no, it does not */
	    rootedAtSubDirListDetailsp -> subDirListUsed = SUBDIRLISTNOTUSED ;

	
	    if( rootedAtSubDirListDetailsp -> subDirListRDN != NULLRDN )
	    {
	    	rdn_free( rootedAtSubDirListDetailsp -> subDirListRDN ) ;
	    	rootedAtSubDirListDetailsp -> subDirListRDN = NULLRDN ;
	    }
	}
	else
	{
	    if( rootedAtSubDirListDetailsp -> subDirListRDN == NULLRDN ||
	        rdn_cmp( rootedAtSubDirListDetailsp -> subDirListRDN, subDirListRDN ) != 0 )
	    {
	    	if( rootedAtSubDirListDetailsp -> subDirListRDN != NULLRDN )
		    rdn_free( rootedAtSubDirListDetailsp -> subDirListRDN ) ;

		rootedAtSubDirListDetailsp -> subDirListRDN = rdn_cpy( subDirListRDN ) ;
	    }
	    
	    /* Do we already have the expanded DN components list.. */
	    if( rootedAtSubDirListDetailsp->subDirListUsed  == SUBDIRLISTNOTUSED ||
	    	rootedAtSubDirListDetailsp->rootedAtDN 	    == NULLDN 	    ||
	    	rootedAtSubDirListDetailsp->bulkSubTreeDN   == NULLDN 	    ||
	    	dn_cmp( rootedAtSubDirListDetailsp->rootedAtDN, rootedAtDN) != 0 )
	    {
	    	/* ..No */
		
		if( rootedAtSubDirListDetailsp->bulkSubTreeDNextensions != ((DN *)NULL) )
		{
		    /* Run through the DN list (which is terminated with a NULLDN).. */
		    
		    baseDNextList = rootedAtSubDirListDetailsp->bulkSubTreeDNextensions ;
		    DNindex = 0 ;
		    
		    while( *(baseDNextList + DNindex) != NULLDN )
		    {
		    	dn_free( *(baseDNextList + DNindex) ) ;
			DNindex++ ;
		    }

		    /* Free the list pointer storage itself.. */
		    free( (char *)baseDNextList ) ;		    
		    
		    rootedAtSubDirListDetailsp->bulkSubTreeDNextensions = ((DN *)NULL) ;
		}

		if( rootedAtSubDirListDetailsp->rootedAtDN != NULLDN )
		{
		    dn_free( rootedAtSubDirListDetailsp->rootedAtDN ) ;
		    rootedAtSubDirListDetailsp->rootedAtDN = NULLDN ;
		}
		
		if( rootedAtSubDirListDetailsp->bulkSubTreeDN != NULLDN )
		{
		    dn_free( rootedAtSubDirListDetailsp->bulkSubTreeDN ) ;
		    rootedAtSubDirListDetailsp->bulkSubTreeDN = NULLDN ;
		}
		
		rootedAtSubDirListDetailsp->rootedAtDN = dn_cpy( rootedAtDN ) ;

		/* Check for the special case Sub directory list name (ALL).. */
		if( (subDirListNamecp = AttrV2pstr( &(subDirListRDN -> rdn_av))) != NULLCP &&
		    strcasecmp( subDirListNamecp, "ALL" ) == 0 )
		{
		    rootedAtSubDirListDetailsp -> subDirListUsed = SUBDIRLISTALLSUBORDSUSED ;

		    /* Remove the "bulkSubDirList" DN component */
		    rootedAtSubDirListDetailsp -> bulkSubTreeDN   = dn_cpy( rootedAtDN ) ;

		    if( (nextToLastRootedAtDN = findNextToLastDN(rootedAtSubDirListDetailsp -> bulkSubTreeDN)) != NULLDN )
		    {
			dn_free( nextToLastRootedAtDN -> dn_parent );
		    	nextToLastRootedAtDN -> dn_parent = NULLDN ;
		    }
		
		    /* NOTE: the ~bulkSubTreeDNextensions" are not set up here as they are
		       retrieved only when necessary e.g. within userInputEntryOrParentDNresolution() */
		}
		else
		{
		    rootedAtSubDirListDetailsp -> subDirListUsed = SUBDIRLISTUSED ;
		    
		    /* Attempt to retrieve the sub directory list entry.. */
		    addToList( &attrsToBeReturnedp, BL_BULK_SUB_TREE_DN  ) ;
		    addToList( &attrsToBeReturnedp, BL_BULK_SUB_TREE_DN_EXT ) ;

		    if( readEntry( rootedAtDN, &readSubDirListp, attrsToBeReturnedp, FALSE ) == OK && readSubDirListp != NULLLIST )
		    {
		    	/* Retrieve the "bulkSubDir" DN attributes.. */
		    	for( as = readSubDirListp->ats ; as != NULLATTR ; as = as->attr_link )
		    	{
			    if( strcmp( attr2name( as->attr_type, OIDPART ), BL_BULK_SUB_TREE_DN ) == 0 )
			    {
			    	/* Convert the the single DN attribute value to a DN.. */
				if( (as->attr_value != NULLAV) &&
				    (DNcp = AttrV2pstr( &(as->attr_value->avseq_av ))) != NULLCP )
				{
				    rootedAtSubDirListDetailsp -> bulkSubTreeDN = str2dn( DNcp ) ;
				    free( DNcp ) ;
				}				
			    }
			    else if( strcmp( attr2name( as->attr_type, OIDPART ), BL_BULK_SUB_TREE_DN_EXT ) == 0 )
			    {
			    	/* Initially count the number of attribute value DNs.. */
			    	for( DNcount=0, avs = as -> attr_value ; avs != NULLAV ; avs = avs -> avseq_next )
			    	    DNcount++ ;

			    	/* Malloc the list of pointers (including one NULLDN terminating pointer).. */
			    	if( (rootedAtSubDirListDetailsp->bulkSubTreeDNextensions = (DN *)malloc( sizeof(DN *) * (DNcount + 1) )) !=  ((DN *)NULL) )
			    	{
			    	    baseDNextList = rootedAtSubDirListDetailsp->bulkSubTreeDNextensions ; 

				    /* Convert each of the attributes values to their DN form.. */
			    	    for( DNindex = 0, avs = as -> attr_value ; avs != NULLAV ; DNindex++, avs = avs -> avseq_next )
			    	    {
				    	if( (DNcp = AttrV2pstr( &(avs->avseq_av ))) != NULLCP )
				    	{
				    	    *(baseDNextList + DNindex) = str2dn( DNcp ) ;
					    free( DNcp ) ;
				    	}
				    }

				    /* Null terminate the list.. */
				    *(baseDNextList + DNindex) = NULLDN ;
				}
			    	else {
				    strcpy( errorDiagnosticscp, "Malloc failure when retrieving the Sub directory list" ) ;
				    retVal = NOTOK ;
				}
			    }
		    	}
		    
		    	freeListpp( &readSubDirListp ) ;
		    }
		    else
 		    {
		    	strcpy( errorDiagnosticscp, "Sub directory list could not be expanded" ) ;
			retVal = NOTOK ;
		    }
		
		    freeListpp( &attrsToBeReturnedp ) ;		
		}
		
		if( subDirListNamecp != NULLCP )
		    free( subDirListNamecp ) ;
	    }
	}

	return( retVal ) ;
}


/*
 * Note: bulkSubTreeDNextensions should equal ((DN *)NULL) i.e. any previous contents
 * must be freed before calling this procedure.
 *
 * Clearly the DN extensions from a LIST operation will be RDNs..
 */

void
setUpTheListAllSubordinateDNextensions( rootedAtSubDirListDetailsp, diagnosticscp )
struct  subDirListDetails * rootedAtSubDirListDetailsp ;
char *	diagnosticscp ; /* A pointer to an DIAGBUFLEN buffer */
{
int	RDNcount = 0, RDNindex = 0 ;
char *	errorMsgcp = NULLCP ;
struct  subordinate * subordinatesp = NULLSUBORD, * currSubordp ;
DN  *	baseDNextList ;
AttributeType	cnat = NULLAttrT, uidat = NULLAttrT, dataSourceat = NULLAttrT, bulkSubDirListat = NULLAttrT ;

	uidat = str2AttrT( BL_UID ) ; cnat = str2AttrT( BL_COMMON_NAME ) ; 
	dataSourceat = str2AttrT( BL_DATA_SOURCE ) ; bulkSubDirListat = str2AttrT( BL_BULK_SUB_DIR_LIST ) ;

	if( listSubtree( rootedAtSubDirListDetailsp->bulkSubTreeDN, &errorMsgcp, &subordinatesp ) == OK )
	{
	    if( subordinatesp == NULLSUBORD )
	    	strcpy( diagnosticscp, "No subordinate DN extensions were retrieved for the \"All\" sub-directory list" ) ;

	    else
	    {
		/* Initially count the number of sub-ordinate RDNs.. */
		for( RDNcount=0, currSubordp = subordinatesp ; currSubordp != NULLSUBORD ; currSubordp = currSubordp -> sub_next )
		    RDNcount++ ;

		/* Malloc the list of pointers (including one NULLDN terminating pointer).. */
		if( (rootedAtSubDirListDetailsp->bulkSubTreeDNextensions = (DN *)malloc( sizeof(DN *) * (RDNcount + 1) )) !=  ((DN *)NULL) )
		{
		    baseDNextList = rootedAtSubDirListDetailsp->bulkSubTreeDNextensions ;
		    
		    for( RDNindex = 0, currSubordp = subordinatesp ; currSubordp != NULLSUBORD ; currSubordp = currSubordp -> sub_next )
		    {
		    	if( currSubordp -> sub_rdn != NULLRDN &&
	    	 	    findAttrTypeRdnComp( cnat,  currSubordp -> sub_rdn )            == NULLRDN &&
	          	    findAttrTypeRdnComp( uidat, currSubordp -> sub_rdn )            == NULLRDN &&
	          	    findAttrTypeRdnComp( dataSourceat, currSubordp -> sub_rdn )     == NULLRDN &&
	          	    findAttrTypeRdnComp( bulkSubDirListat, currSubordp -> sub_rdn ) == NULLRDN   )
			{
			    *(baseDNextList + RDNindex) = dn_comp_new( currSubordp -> sub_rdn ) ;
			    RDNindex++ ;
			    currSubordp -> sub_rdn = NULLRDN ; /* To prevent it from being freed ! */
			}
		    }
		    
		    /* Null terminate the list.. */
		    *(baseDNextList + RDNindex) = NULLDN ;
		}
		else strcpy( diagnosticscp, "Memory allocation failure in setUpTheListAllSubordinateDNextensions()" ) ;
		
	    	free_subordinatepp( &subordinatesp ) ;
	    }
	}
	else
	{
	    if( errorMsgcp != NULLCP )
	    {
	    	strcpy( diagnosticscp, errorMsgcp ) ;
		free( errorMsgcp ) ;
	    }
	}

	AttrT_free( cnat ) ; AttrT_free( uidatat ) ; AttrT_free( dataSourceat ) ; AttrT_free( bulkSubDirListat ) ;

	return ;
}
